// Copied from https://github.com/RaymiiOrg/cpp-command-output.git  9a9edca13f15182b0c405902747f86c4a19638ec
#pragma once
// Copyright (C) 2021 Remy van Elst
//
//     This program is free software: you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation, either version 3 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
#include <array>
#include <ostream>
#include <string>
#include <cstdio>

#ifdef __FreeBSD__
#include <sys/wait.h>
#endif

namespace raymii {

    struct CommandResult {
        friend std::ostream& operator << (std::ostream& os, const CommandResult& result) {
            os << "command exitstatus: " << result.exitstatus << " output: " << result.output;
            return os;
        }

        bool operator == (const CommandResult& rhs) const {
            return output == rhs.output && exitstatus == rhs.exitstatus;
        }

        bool operator != (const CommandResult& rhs) const {
            return !(rhs == *this);
        }

        std::string output;
        int exitstatus;
    };

    class Command {
    public:
        /**
             * Execute system command and get STDOUT result.
             * Regular system() only gives back exit status, this gives back output as well.
             * @param command system command to execute
             * @return commandResult containing STDOUT (not stderr) output & exitstatus
             * of command. Empty if command failed (or has no output). If you want stderr,
             * use shell redirection (2&>1).
             */
        static CommandResult exec(const std::string &command) {
            int exitcode = 0;
            std::array<char, 8192> buffer{};
            std::string result;
#ifdef _WIN32
#define popen _popen
#define pclose _pclose
#define WEXITSTATUS
#endif
            FILE *pipe = popen(command.c_str(), "r");
            if (pipe == nullptr) {
                throw std::runtime_error("popen() failed!");
            }
            try {
                std::size_t bytesread;
                while ((bytesread = std::fread(buffer.data(), sizeof(buffer.at(0)), sizeof(buffer), pipe)) != 0) {
                    result += std::string(buffer.data(), bytesread);
                }
            } catch (...) {
                pclose(pipe);
                throw;
            }
            // Workaround "error: cannot take the address of an rvalue of type 'int'" on MacOS
            // see e.g. https://github.com/BestImageViewer/geeqie/commit/75c7df8b96592e10f7936dc1a28983be4089578c
            int res = pclose(pipe);
            exitcode = WEXITSTATUS(res);
            return CommandResult{result, exitcode};
        }

        // Only for reference in the article. Use regular ::exec.
        static CommandResult execFgets(const std::string &command) {
            int exitcode = 0;
            std::array<char, 8192> buffer{};
            std::string result;
#ifdef _WIN32
#define popen _popen
#define pclose _pclose
#define WEXITSTATUS
#endif
            FILE *pipe = popen(command.c_str(), "r");
            if (pipe == nullptr) {
                throw std::runtime_error("popen() failed!");
            }
            try {
                while (std::fgets(buffer.data(), buffer.size(), pipe) != nullptr) {
                    result += buffer.data();
                }
            } catch (...) {
                pclose(pipe);
                throw;
            }
            // Workaround "error: cannot take the address of an rvalue of type 'int'" on MacOS
            // see e.g. https://github.com/BestImageViewer/geeqie/commit/75c7df8b96592e10f7936dc1a28983be4089578c
            int res = pclose(pipe);
            exitcode = WEXITSTATUS(res);
            return CommandResult{result, exitcode};
        }
    };

}

